// Code generated by "stringer -type=lspTYPE"; DO NOT EDIT.

package zebra

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[lspNone-0]
	_ = x[lspStatic-1]
	_ = x[lspLDP-2]
	_ = x[lspBGP-3]
	_ = x[lspSR-4]
	_ = x[lspSHARP-5]
}

const _lspTYPE_name = "lspNonelspStaticlspLDPlspBGPlspSRlspSHARP"

var _lspTYPE_index = [...]uint8{0, 7, 16, 22, 28, 33, 41}

func (i lspTYPE) String() string {
	if i >= lspTYPE(len(_lspTYPE_index)-1) {
		return "lspTYPE(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _lspTYPE_name[_lspTYPE_index[i]:_lspTYPE_index[i+1]]
}
