// Code generated by "stringer -type=RouteType"; DO NOT EDIT.

package zebra

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[routeSystem-0]
	_ = x[routeKernel-1]
	_ = x[routeConnect-2]
	_ = x[RouteStatic-3]
	_ = x[routeRIP-4]
	_ = x[routeRIPNG-5]
	_ = x[routeOSPF-6]
	_ = x[routeOSPF6-7]
	_ = x[routeISIS-8]
	_ = x[RouteBGP-9]
	_ = x[routePIM-10]
	_ = x[routeEIGRP-11]
	_ = x[routeNHRP-12]
	_ = x[routeHSLS-13]
	_ = x[routeOLSR-14]
	_ = x[routeTABLE-15]
	_ = x[routeLDP-16]
	_ = x[routeVNC-17]
	_ = x[routeVNCDirect-18]
	_ = x[routeVNCDirectRH-19]
	_ = x[routeBGPDirect-20]
	_ = x[routeBGPDirectEXT-21]
	_ = x[routeBABEL-22]
	_ = x[routeSHARP-23]
	_ = x[routePBR-24]
	_ = x[routeBFD-25]
	_ = x[routeOpenfabric-26]
	_ = x[routeVRRP-27]
	_ = x[routeNHG-28]
	_ = x[routeSRTE-29]
	_ = x[routeAll-30]
	_ = x[routeMax-31]
	_ = x[zapi5Frr4RouteAll-24]
	_ = x[zapi5Frr5RouteAll-25]
	_ = x[zapi6Frr6RouteAll-26]
	_ = x[zapi6Frr7RouteAll-27]
	_ = x[zapi6Frr7dot2RouteAll-28]
	_ = x[zapi6Frr7dot3RouteAll-29]
	_ = x[zapi4RouteNHRP-11]
	_ = x[zapi4RouteHSLS-12]
	_ = x[zapi4RouteOLSR-13]
	_ = x[zapi4RouteTABLE-14]
	_ = x[zapi4RouteLDP-15]
	_ = x[zapi4RouteVNC-16]
	_ = x[zapi4RouteVNCDirect-17]
	_ = x[zapi4RouteVNCDirectRH-18]
	_ = x[zapi4RouteBGPDixrect-19]
	_ = x[zapi4RouteBGPDirectEXT-20]
	_ = x[zapi4RouteAll-21]
	_ = x[zapi3RouteHSLS-11]
	_ = x[zapi3RouteOLSR-12]
	_ = x[zapi3RouteBABEL-13]
	_ = x[zapi3RouteNHRP-14]
}

const _RouteType_name = "routeSystemrouteKernelrouteConnectRouteStaticrouteRIProuteRIPNGrouteOSPFrouteOSPF6routeISISRouteBGProutePIMrouteEIGRProuteNHRProuteHSLSrouteOLSRrouteTABLErouteLDProuteVNCrouteVNCDirectrouteVNCDirectRHrouteBGPDirectrouteBGPDirectEXTrouteBABELrouteSHARProutePBRrouteBFDrouteOpenfabricrouteVRRProuteNHGrouteSRTErouteAllrouteMax"

var _RouteType_index = [...]uint16{0, 11, 22, 34, 45, 53, 63, 72, 82, 91, 99, 107, 117, 126, 135, 144, 154, 162, 170, 184, 200, 214, 231, 241, 251, 259, 267, 282, 291, 299, 308, 316, 324}

func (i RouteType) String() string {
	if i >= RouteType(len(_RouteType_index)-1) {
		return "RouteType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _RouteType_name[_RouteType_index[i]:_RouteType_index[i+1]]
}
