class _ParserOption:
    test_prefix = ""
    gobgp_image = "osrg/gobgp"
    exabgp_path = ""
    go_path = ""
    gobgp_log_level = "info"
    test_index = 0
    config_format = "yaml"


parser_option = _ParserOption()
