# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    Bridge,
    BGP_FSM_ESTABLISHED,
    local,
)
from lib.gobgp import GoBGPContainer
from lib.quagga import QuaggaBGPContainer


class GoBGPTestBase(unittest.TestCase):
    # The same topology for IPv4 and IPv6 environment.
    # o1: Quagga (not router)
    # g1: GoBGP
    # q1: Quagga (Zebra + BGPd)
    # o2: Quagga (not router)
    #
    # +----+       +----+           +----+       +----+
    # | o1 |       | g1 |===(BGP)===| q1 |       | o2 |
    # +----+       +----+           +----+       +----+
    #    |          |  |             |  |          |
    #    +--(br01)--+  +---(br02)----+  +--(br03)--+

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        # preparing the container for ipv4
        g1_v4 = GoBGPContainer(name='g1_v4', asn=65000, router_id='192.168.0.1',
                               ctn_image_name=gobgp_ctn_image_name,
                               log_level=parser_option.gobgp_log_level,
                               zebra=True)
        q1_v4 = QuaggaBGPContainer(name='q1_v4', asn=65001, router_id='192.168.0.2', zebra=True)
        o1_v4 = QuaggaBGPContainer(name='o1_v4', asn=65002, router_id='192.168.0.3')
        o2_v4 = QuaggaBGPContainer(name='o2_v4', asn=65002, router_id='192.168.0.4')

        # preparing the container for ipv6
        g1_v6 = GoBGPContainer(name='g1_v6', asn=65000, router_id='192.168.0.1',
                               ctn_image_name=gobgp_ctn_image_name,
                               log_level=parser_option.gobgp_log_level,
                               zebra=True)
        q1_v6 = QuaggaBGPContainer(name='q1_v6', asn=65001, router_id='192.168.0.2', zebra=True)
        o1_v6 = QuaggaBGPContainer(name='o1_v6', asn=65002, router_id='192.168.0.3')
        o2_v6 = QuaggaBGPContainer(name='o2_v6', asn=65002, router_id='192.168.0.4')

        # preparing the bridge for ipv4
        br01_v4 = Bridge(name='br01_v4', subnet='192.168.10.0/24')
        br02_v4 = Bridge(name='br02_v4', subnet='192.168.20.0/24')
        br03_v4 = Bridge(name='br03_v4', subnet='192.168.30.0/24')

        # preparing the bridge for ipv6
        br01_v6 = Bridge(name='br01_v6', subnet='2001:10::/32')
        br02_v6 = Bridge(name='br02_v6', subnet='2001:20::/32')
        br03_v6 = Bridge(name='br03_v6', subnet='2001:30::/32')

        cls.ctns = {
            'ipv4': [g1_v4, q1_v4, o1_v4, o2_v4],
            'ipv6': [g1_v6, q1_v6, o1_v6, o2_v6],
        }
        cls.gobgps = {
            'ipv4': g1_v4,
            'ipv6': g1_v6,
        }
        cls.quaggas = {
            'ipv4': q1_v4,
            'ipv6': q1_v6,
        }
        cls.others = {
            'ipv4': [o1_v4, o2_v4],
            'ipv6': [o1_v6, o2_v6],
        }
        cls.bridges = {
            'br01_v4': br01_v4,
            'br02_v4': br02_v4,
            'br03_v4': br03_v4,
            'br01_v6': br01_v6,
            'br02_v6': br02_v6,
            'br03_v6': br03_v6,
        }

    """
      No.1 start up ipv4 containers and check state
           each neighbor is established in ipv4 environment
    """

    def test_01_check_neighbor_established(self):
        g1 = self.gobgps['ipv4']
        q1 = self.quaggas['ipv4']
        o1 = self.others['ipv4'][0]
        o2 = self.others['ipv4'][1]
        # start up containers of ipv4 environment
        initial_wait_time = max(ctn.run() for ctn in self.ctns['ipv4'])
        time.sleep(initial_wait_time)

        # make ipv4 bridge and set ip to each container
        [self.bridges['br01_v4'].addif(ctn) for ctn in [o1, g1]]
        [self.bridges['br02_v4'].addif(ctn) for ctn in [g1, q1]]
        [self.bridges['br03_v4'].addif(ctn) for ctn in [q1, o2]]

        g1.add_peer(q1, bridge=self.bridges['br02_v4'].name)
        q1.add_peer(g1, bridge=self.bridges['br02_v4'].name)

        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q1)

    """
      No.2 check whether the ping is reachable in container
           that have previously beyond the GoBGP in ipv4 environment
    """

    def test_02_check_reachability_between_o1_and_o2(self):
        g1 = self.gobgps['ipv4']
        q1 = self.quaggas['ipv4']
        o1 = self.others['ipv4'][0]
        o2 = self.others['ipv4'][1]

        # set o1's default gateway as g1
        g1_addr = g1.ip_addrs[1][1].split('/')[0]
        o1.add_static_route(self.bridges['br03_v4'].subnet, g1_addr)

        # set o2's default gateway as q1
        q1_addr = q1.ip_addrs[2][1].split('/')[0]
        o2.add_static_route(self.bridges['br01_v4'].subnet, q1_addr)

        # test reachability between o1 and o2
        addrs = [e[1] for e in o2.ip_addrs if 'br03_v4' in e[2]]
        self.assertEqual(len(addrs), 1)
        o2_addr = addrs[0]
        o1.get_reachability(o2_addr)

    """
      No.4 start up ipv6 containers and check state
           each neighbor is established in ipv6 environment
    """

    def test_04_check_neighbor_established_v6(self):
        g1 = self.gobgps['ipv6']
        q1 = self.quaggas['ipv6']
        o1 = self.others['ipv6'][0]
        o2 = self.others['ipv6'][1]
        # start up containers of ipv6 environment
        initial_wait_time = max(ctn.run() for ctn in self.ctns['ipv6'])
        time.sleep(initial_wait_time)

        # make ipv6 bridge and set ip to each container
        [self.bridges['br01_v6'].addif(ctn) for ctn in [o1, g1]]
        [self.bridges['br02_v6'].addif(ctn) for ctn in [g1, q1]]
        [self.bridges['br03_v6'].addif(ctn) for ctn in [q1, o2]]

        g1.add_peer(q1, bridge=self.bridges['br02_v6'].name, v6=True)
        q1.add_peer(g1, bridge=self.bridges['br02_v6'].name, v6=True)

        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q1)

    """
      No.5 check whether the ping is reachable in container
           that have previously beyond the GoBGP in ipv6 environment
    """

    def test_05_check_reachability_between_o1_and_o2(self):
        g1 = self.gobgps['ipv6']
        q1 = self.quaggas['ipv6']
        o1 = self.others['ipv6'][0]
        o2 = self.others['ipv6'][1]

        # set o1's default gateway as g1
        g1_addr = [a for a in g1.ip6_addrs if a[2] == self.bridges['br01_v6'].name][0][1].split('/')[0]
        o1.add_static_route(self.bridges['br03_v6'].subnet, g1_addr)

        # set o2's default gateway as q1
        q1_addr = [a for a in q1.ip6_addrs if a[2] == self.bridges['br03_v6'].name][0][1].split('/')[0]
        o2.add_static_route(self.bridges['br01_v6'].subnet, q1_addr)

        # test reachability between o1 and o2
        addrs = [e[1] for e in o2.ip6_addrs if 'br03_v6' in e[2]]
        self.assertEqual(len(addrs), 1)
        o2_addr = addrs[0]
        o1.get_reachability(o2_addr)

    def test_07_mpath_test_setup(self):
        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=parser_option.gobgp_image,
                            log_level=parser_option.gobgp_log_level,
                            config_format=parser_option.config_format,
                            zebra=True)
        g2 = GoBGPContainer(name='g2', asn=65001, router_id='192.168.0.2',
                            ctn_image_name=parser_option.gobgp_image)
        g3 = GoBGPContainer(name='g3', asn=65001, router_id='192.168.0.3',
                            ctn_image_name=parser_option.gobgp_image)
        g4 = GoBGPContainer(name='g4', asn=65000, router_id='192.168.0.4',
                            ctn_image_name=parser_option.gobgp_image)
        g5 = GoBGPContainer(name='g5', asn=65000, router_id='192.168.0.5',
                            ctn_image_name=parser_option.gobgp_image)

        ctns = [g1, g2, g3, g4, g5]
        for ctn in ctns:
            self.ctns[ctn.name] = ctn

        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time)

        # advertise same prefix
        g2.add_route('10.0.10.0/24')
        g3.add_route('10.0.10.0/24')
        g4.add_route('10.0.10.0/24')
        g5.add_route('10.0.10.0/24')

        for g in [g2, g3, g4, g5]:
            g1.add_peer(g)
            g.add_peer(g1)

    def test_08_mpath_test_check_neighbor_established(self):
        g1 = self.ctns['g1']
        g2 = self.ctns['g2']
        g3 = self.ctns['g3']
        g4 = self.ctns['g4']
        g5 = self.ctns['g5']
        for g in [g2, g3, g4, g5]:
            g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g)

    def test_09_mpath_test_check_mpath_injected(self):
        g1 = self.ctns['g1']
        g2 = self.ctns['g2']
        g3 = self.ctns['g3']
        g4 = self.ctns['g4']
        g5 = self.ctns['g5']

        def nexthops():
            n = []
            for line in g1.local('ip route show 10.0.10.0/24', capture=True).split('\n'):
                line = line.strip()
                if 'via' in line:
                    n.append(line.split(' ')[2].strip())
            return n

        def validate_nexthops(peers):
            interval = 1
            count = 0
            timeout = 30
            while True:
                valid = False
                nhs = nexthops()
                if len(nhs) == len(peers):
                    valid = True
                    for peer in peers:
                        if g1.peers[peer]['neigh_addr'].split('/')[0] not in nhs:
                            valid = False
                            break
                if valid:
                    return

                time.sleep(interval)
                count += interval
                if count >= timeout:
                    raise Exception(nhs)

        validate_nexthops([g4, g5])

        g4.local('gobgp g ri del 10.0.10.0/24')
        validate_nexthops([g5])

        g4.local('gobgp g ri add 10.0.10.0/24 local-pref 200')
        validate_nexthops([g4])

        g4.local('gobgp g ri del 10.0.10.0/24')
        g5.local('gobgp g ri del 10.0.10.0/24')
        validate_nexthops([g2, g3])

        g3.local('gobgp g ri del 10.0.10.0/24')
        validate_nexthops([g2])

        g3.local('gobgp g ri add 10.0.10.0/24 med 10')
        validate_nexthops([g2])

        g2.local('gobgp g ri add 10.0.10.0/24 med 20')
        validate_nexthops([g3])


