// Code generated by "stringer -type=ptmStatus"; DO NOT EDIT.

package zebra

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ptmStatusDown-0]
	_ = x[ptmStatusUp-1]
	_ = x[ptmStatusUnknown-2]
}

const _ptmStatus_name = "ptmStatusDownptmStatusUpptmStatusUnknown"

var _ptmStatus_index = [...]uint8{0, 13, 24, 40}

func (i ptmStatus) String() string {
	if i >= ptmStatus(len(_ptmStatus_index)-1) {
		return "ptmStatus(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ptmStatus_name[_ptmStatus_index[i]:_ptmStatus_index[i+1]]
}
