// Code generated by "stringer -type=afi"; DO NOT EDIT.

package zebra

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[afiIP-1]
	_ = x[afiIP6-2]
	_ = x[afiEther-3]
	_ = x[afiMax-4]
}

const _afi_name = "afiIPafiIP6afiEtherafiMax"

var _afi_index = [...]uint8{0, 5, 11, 19, 25}

func (i afi) String() string {
	i -= 1
	if i >= afi(len(_afi_index)-1) {
		return "afi(" + strconv.FormatInt(int64(i+1), 10) + ")"
	}
	return _afi_name[_afi_index[i]:_afi_index[i+1]]
}
