// Generated code; DO NOT EDIT.

package bgp

const (
	RESERVED              SRBehavior = 0
	END                   SRBehavior = 1
	END_WITH_PSP          SRBehavior = 2
	END_WITH_USP          SRBehavior = 3
	END_WITH_PSP_USP      SRBehavior = 4
	ENDX                  SRBehavior = 5
	ENDX_WITH_PSP         SRBehavior = 6
	ENDX_WITH_USP         SRBehavior = 7
	ENDX_WITH_PSP_USP     SRBehavior = 8
	ENDT                  SRBehavior = 9
	ENDT_WITH_PSP         SRBehavior = 10
	ENDT_WITH_USP         SRBehavior = 11
	ENDT_WITH_PSP_USP     SRBehavior = 12
	END_B6_ENCAPS         SRBehavior = 14
	END_BM                SRBehavior = 15
	END_DX6               SRBehavior = 16
	END_DX4               SRBehavior = 17
	END_DT6               SRBehavior = 18
	END_DT4               SRBehavior = 19
	END_DT46              SRBehavior = 20
	END_DX2               SRBehavior = 21
	END_DX2V              SRBehavior = 22
	END_DT2U              SRBehavior = 23
	END_DT2M              SRBehavior = 24
	END_B6_ENCAPS_Red     SRBehavior = 27
	END_WITH_USD          SRBehavior = 28
	END_WITH_PSP_USD      SRBehavior = 29
	END_WITH_USP_USD      SRBehavior = 30
	END_WITH_PSP_USP_USD  SRBehavior = 31
	ENDX_WITH_USD         SRBehavior = 32
	ENDX_WITH_PSP_USD     SRBehavior = 33
	ENDX_WITH_USP_USD     SRBehavior = 34
	ENDX_WITH_PSP_USP_USD SRBehavior = 35
	ENDT_WITH_USD         SRBehavior = 36
	ENDT_WITH_PSP_USD     SRBehavior = 37
	ENDT_WITH_USP_USD     SRBehavior = 38
	ENDT_WITH_PSP_USP_USD SRBehavior = 39
	ENDM_GTP6D            SRBehavior = 69 // 0x0045
	ENDM_GTP6DI           SRBehavior = 70 // 0x0046
	ENDM_GTP6E            SRBehavior = 71 // 0x0047
	ENDM_GTP4E            SRBehavior = 72 // 0x0048
)
