// Code generated by "stringer -type=DiagnosticType,StateType -linecomment -output=bfd_types_string.go"; DO NOT EDIT.

package bfd

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[DiagnosticNoDiagnostic-0]
	_ = x[DiagnosticControlDetectionTimeExpired-1]
	_ = x[DiagnosticEchoFunctionFailed-2]
	_ = x[DiagnosticNeighborSignaledSessionDown-3]
	_ = x[DiagnosticForwardingPlaneReset-4]
	_ = x[DiagnosticPathDown-5]
	_ = x[DiagnosticConcatenatedPathDown-6]
	_ = x[DiagnosticAdministrativelyDown-7]
	_ = x[DiagnosticReverseConcatenatedPathDown-8]
	_ = x[DiagnosticReservedStart-9]
	_ = x[DiagnosticReservedEnd-31]
}

const (
	_DiagnosticType_name_0 = "no diagnosticcontrol detection time expiredecho function failedneighbor signaled session downforwarding plane resetpath downconcatenated path downadministratively downreverse concatenated path downreserved (9-31)"
	_DiagnosticType_name_1 = "reserved (9-31)"
)

var (
	_DiagnosticType_index_0 = [...]uint8{0, 13, 43, 63, 93, 115, 124, 146, 167, 197, 212}
)

func (i DiagnosticType) String() string {
	switch {
	case i <= 9:
		return _DiagnosticType_name_0[_DiagnosticType_index_0[i]:_DiagnosticType_index_0[i+1]]
	case i == 31:
		return _DiagnosticType_name_1
	default:
		return "DiagnosticType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
}
func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[StateAdminDown-0]
	_ = x[StateDown-1]
	_ = x[StateInit-2]
	_ = x[StateUp-3]
}

const _StateType_name = "admin downdowninitup"

var _StateType_index = [...]uint8{0, 10, 14, 18, 20}

func (i StateType) String() string {
	idx := int(i) - 0
	if i < 0 || idx >= len(_StateType_index)-1 {
		return "StateType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _StateType_name[_StateType_index[idx]:_StateType_index[idx+1]]
}
