// Code generated by "stringer -type=ptmEnable"; DO NOT EDIT.

package zebra

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ptmEnableOff-0]
	_ = x[ptmEnableOn-1]
	_ = x[ptmEnableUnspec-2]
}

const _ptmEnable_name = "ptmEnableOffptmEnableOnptmEnableUnspec"

var _ptmEnable_index = [...]uint8{0, 12, 23, 38}

func (i ptmEnable) String() string {
	if i >= ptmEnable(len(_ptmEnable_index)-1) {
		return "ptmEnable(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ptmEnable_name[_ptmEnable_index[i]:_ptmEnable_index[i+1]]
}
