// Code generated by "stringer -type=Safi"; DO NOT EDIT.

package zebra

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[safiUnspec-0]
	_ = x[SafiUnicast-1]
	_ = x[safiMulticast-2]
	_ = x[safiMplsVpn-3]
	_ = x[safiEncap-4]
	_ = x[safiEvpn-5]
	_ = x[safiLabeledUnicast-6]
	_ = x[safiFlowspec-7]
	_ = x[safiMax-8]
	_ = x[zapi4SafiMplsVpn-3]
	_ = x[zapi3SafiMplsVpn-4]
	_ = x[zapi4SafiEncap-5]
	_ = x[zapi4SafiEvpn-6]
	_ = x[zapi3SafiEncap-7]
}

const _Safi_name = "safiUnspecSafiUnicastsafiMulticastsafiMplsVpnsafiEncapsafiEvpnsafiLabeledUnicastsafiFlowspecsafiMax"

var _Safi_index = [...]uint8{0, 10, 21, 34, 45, 54, 62, 80, 92, 99}

func (i Safi) String() string {
	if i >= Safi(len(_Safi_index)-1) {
		return "Safi(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _Safi_name[_Safi_index[i]:_Safi_index[i+1]]
}
