# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from lib import base
from lib import utils
from lib.base import BGP_FSM_ESTABLISHED, local, Bridge
from lib.gobgp import GoBGPContainer
import sys
import os
import time

import collections
collections.Callable = collections.abc.Callable

from lib.noseplugin import parser_option
from itertools import combinations


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g2 = GoBGPContainer(name='g2', asn=65001, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        ctns = [g1, g2]

        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time + 2)

        br = Bridge(name='unnumbered', subnet='fd00:1::/64')
        [br.addif(ctn) for ctn in [g1, g2]]

        utils.probe_link_local_address(g1, g2, 'eth1', 'eth1')

        g1.add_peer(g2, interface='eth1')
        g2.add_peer(g1, interface='eth1')

        cls.g1 = g1
        cls.g2 = g2

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        self.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)
        self.g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)

    def test_02_add_ipv4_route(self):

        self.g1.add_route('10.0.0.0/24')

        time.sleep(1)

        rib = self.g2.get_global_rib(rf='ipv4')
        self.assertEqual(len(rib), 1)


